#!/usr/bin/perl

# exiscan 0.6 by tom kistner <tom@duncanthrax.net>

# to make exiscan exit gracefully, send it a HUP signal
# see INSTALL for installation instructions

# ******************************************************************
# ------------------------------------------------------------------
# Configuration Section - edit here !

# this is where you have put the exiscan package
$basepath="/usr/exim/exiscan";

# exims queue directory
$queuedir="/usr/exim/spool/input";

# the metamail executable. used to unpack mails
$metamail="/usr/bin/metamail";

# metasend, to forward the virus infected mails as an attach to
# the admin
$metasend="/usr/bin/metasend";

# the virus scanner to use ("mcafee" or "trendmicro")
$scanner="trendmicro";

# the virus scanner executable (full path + executable)
$scannerex="/etc/iscan/vscan";

# the exim executable
$exim="/usr/exim/bin/exim";

# exiscan "forgets" about already scanned mail
# ID's every $resetinterval seconds. This should
# be set to the value of the exim dequeue interval
# (if you ran exim with a flag of -q1h, then you
# should set "3600" here. 3600 secs = 1hr)
$resetinterval="3600";

# the email addy that gets notification about found
# viruses. mind the escaped '@' !
$postmaster="postmaster\@eurescom.de";

# the syslog facility to use. leave it as it is to
# get outpuy in /var/log/messages (on most machines)
$facility = "LOG_EXISCAN";

# ------------------------------------------------------------------
# ******************************************************************



# -------------------------------------------
# you may adjust these settings too
# but it's not necessary if you use
# the standard exiscan directory layout

# directory where mails are unpacked
$checkdir="$basepath/checkqueue";

# where to put virus infested mails
$virusdir="$basepath/virusmails";

# -------------------------------------------


# -------------------------------------------
# no editing below this line !

use File::Copy;
use File::Remove qw(remove);

# now some sanity checks

print "Checking for metamail .. ";
if (!(-x $metamail)) {
	print "not found ($metamail) ! Please check your settings\n";
	exit(1);
};
print "found\n";

print "Checking for metasend .. ";
if (!(-x $metasend)) {
        print "not found ($metasend) ! Please check your settings\n";
        exit(1);
};
print "found\n";

print "Checking for virus scanner .. ";
if (!(-x $scannerex)) {
        print "not found ($scannerex) ! Please check your settings\n";
        exit(1);
};
print "found\n";

print "Checking for exim .. ";
if (!(-x $exim)) {
        print "not found ($exim) ! Please check your settings\n";
        exit(1);
};
print "found\n";

$quit = 0;
$| = 0;

# fork a child and exit
FORK: {
	if ($pid = fork) {
		print "exiscan launched into the background (pid $pid)\n";
		exit(0);
	}
	elsif (defined $pid) {
		goto CHILD;
	}
	else {
		die "exiscan can't fork to the background: $!\n";
	};
};

CHILD:

%checked = ();
$lastreset = time();

# ---------------------------------
# the exit handler
# HUP signal makes exiscan quit
# gracefully

sub catch_zap {
       my $signame = shift;
       $quit = 1;        
}

$SIG{HUP} = \&catch_zap;

# --------------------------------------
# set up syslog

use Unix::Syslog qw(:macros);
use Unix::Syslog qw(:subs);

openlog "exiscan", LOG_PERROR | LOG_PID, $facility;

syslog LOG_INFO, "started";


# --------------------------------------
# the superloop

while(!($quit)) {

# --------------------------------------
# forget scanned mails IDs every interval
# to allow re-dequeueing

 if ((time() - $lastreset) > $resetinterval) {
 	 syslog LOG_INFO, " exiscan clearing ID cache\n";
	 $lastreset = time();
	 %checked = ();
 };

# --------------------------------------
# get the contents of the queue (headers)

 opendir(QUEUEDIR,"$queuedir");
 @mails = grep /\-H/, readdir QUEUEDIR;
 closedir(QUEUEDIR);

# --------------------------------------
# once second grace to decrease load
# and allow exim to perhaps finish
# writing a header file

 sleep 1;

# ---------------------------------------
# process all found mails

 foreach $mail (@mails) {

	if ($quit) { goto EXITEXISCAN };

	$id = $mail; $id =~ s/\-H$//g;

	if ( (-e "$queuedir/$id-H") && (!($checked{$id})) ) {

               # -------------------------------------
               # check for X-Scanner in the header

		open(HEADER,"< $queuedir/$id-H");
		if ((@scanner = grep /exiscan/i, <HEADER>)) {

			close(HEADER);
			
			($rest,$oldid,$rest2) = split /\*/, @scanner[0], 3;
			if ($checked{$oldid}) {
				syslog LOG_INFO, "ID $id R:resent, prev ID $oldid, dequeuing";
			}
			else {				
				syslog LOG_INFO, "ID $id R:delayed, attempting dequeue again";
			};

			$checked{$id} = 1;
			
                       # ------------------------------------
                       # de-queue 
                      
                        $idplus = $id; $idplus =~ s/.$/9/g;
                        system("$exim -q $id $idplus &");

                       # ------------------------------------
                       # process next mail

			next;
			
		};

		close(HEADER);

	       # -------------------------------------
	       # build complete mail in standard format
	       # get real TO and FROM addresses
	       
		@completemail = ();
		@to = ();
		$i = "";
		
		open(HEADER,"< $queuedir/$id-H");
		do {
			if ($i =~ /^\</) { $from = $i; chop($from); };		
			if ( ($i =~ /.+\@.+/) && (!($i =~ /^\</))) {
				$tmp = $i; chop($tmp);
				push @to, $tmp;
			};
			$i = <HEADER>;
		} while (!($i eq "\n"));
		
		while(<HEADER>) {
			$_ =~ s/^.....//g if ($_ =~ /^[0-9][0-9][0-9]/);
			push @completemail, $_;
		};
		close(HEADER);

		# add our tag
		push @completemail, "X-Scanner: exiscan \*$id\* http://duncanthrax.net/exiscan/\n";

		# add separator
		push @completemail, "\n";

		open(BODY,"< $queuedir/$id-D");
                $i = <BODY>;	# kick out 1st line
		push @completemail, <BODY>;
                close(BODY);

		open(TMPMAIL,"> $checkdir/$id-complete");
		print TMPMAIL @completemail;
		close(TMPMAIL);

               # -------------------------------------
               # unpack mail in temporary directory

		mkdir "$checkdir/$id-tmp", 0755;
		system("METAMAIL_TMPDIR=$checkdir/$id-tmp; export METAMAIL_TMPDIR; $metamail -r -q -x -w $checkdir/$id-complete > /dev/null 2> /dev/null");

               # -------------------------------------
               # run virus scanner and get result

		$result = 0;

		if ($scanner =~ /mcafee/i) {
			$result = system("$scannerex -r $checkdir/$id-tmp > /dev/null 2> /dev/null");
		};

		if ($scanner =~ /trendmicro/i) {
			open(VSCAN,"$scannerex $checkdir/$id-tmp |");

			while(<VSCAN>) {
			        if ($_ =~ /Infected \: /) {
		        	        chop;
		                	($dummy,$infected) = split / \: /,$_,2;
			                if ($infected =~ /[1-9]/) {
		        	                $result = 1;
        		        	        last;
		        	        };
			        };
			};
			close(VSCAN);
		};

		if ($result) {
                
		       # ------------------------------------
                       # Virus found
			
			syslog LOG_INFO, "ID $id F:$from T:@to R:virus found, moving to $virusdir";
	
			move("$queuedir/$id-H","$virusdir/$id-H");
			move("$queuedir/$id-D","$virusdir/$id-D");


		       # ------------------------------------
		       # send notification to admin

			syslog LOG_INFO, "ID $id sending notification to $postmaster";
			system("$metasend -t $postmaster -F exiscan -s \"Virus in mail $id\" -m message/rfc822 -f $checkdir/$id-complete -S 99999999 -b -e 7bit");
		}
		else {
		
		       # ------------------------------------
		       # insert X-Scanner header line	

			open(HEADER,">> $queuedir/$id-H");
			print HEADER "070  X-Scanner: exiscan \*$id\* http://duncanthrax.net/exiscan/\n";
			close(HEADER);

                       # ------------------------------------
                       # remember mail id

			$checked{$id} = 1;

                       # ------------------------------------
                       # Mail is ok, de-queue

			syslog LOG_INFO, "ID $id F:$from T:@to R:mail clean, dequeuing"; 
			system("$exim -Mc $id &");
		};

               # -------------------------------------
               # clean up temp files

		remove \1, "$checkdir/$id-tmp";
		unlink "$checkdir/$id-complete";
	};
   };
 };

# -----------------------------------
# the exit routine. gets passed
# when we shut down with SIGHUP recvd

EXITEXISCAN:

syslog LOG_INFO, "terminating";
closelog;
